#include "compile.h"
#include "machine_state.h"
#include "machine_state_struct.h"
#include "regex_struct.h"

#include "../utf8/utf-8.h"

#include <ctype.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#define COMPILE_STATE_STACK_SIZE 16

static inline bool _IsSpecialChar(int character)
{
	return character == '(' ||
		character == ')' ||
		character == '[' ||
		character == ']' ||
		character == '|';
}

static inline bool _IsQuantifierChar(int character)
{
	return character == '?' ||
		character == '{' ||
		character == '+' ||
		character == '*';
}

static inline bool _IsLiteralChar(int character)
{
	return ! (_IsQuantifierChar(character) || _IsSpecialChar(character) || character == '\\');
}

static inline bool SingleCharacterQuantified(StringView source)
{
	const char* next = UTF8_NextCodepoint(source.source);
	uintptr_t distance = (uintptr_t) next - (uintptr_t) source.source;

	return distance + 1 < source.length
		&& _IsQuantifierChar(source.source[distance]);
}

static RegexMachineStateBase* _NewAcceptState(Regex* regex, StringView* source, size_t length)
{
	// Create new regex state
	RegexMachineStateAccept* accept = Scratchpad_Reserve(
		&regex->machine_memory,
		sizeof(RegexMachineStateAccept)
	);
	if (accept == NULL) {
		return NULL;
	}
	RegexMachineStateAccept_Init(accept);


	// Copy the sequence to an owned buffer
	char* accept_sequence = Scratchpad_Reserve(
		&regex->machine_memory,
		length
	);
	if (accept_sequence == NULL) {
		return NULL;
	}
	memcpy(accept_sequence, source->source, length);

	// Make it a StringView for easier operation, also omits the \0
	accept->sequence = StringView_FromStringSized(
		accept_sequence,
		length
	);

	// Advance source stringview
	*source = StringView_Slice(*source, length, source->length);

	return &accept->base;
}

// Literals are everything except special groups, escaped characters etc...
static RegexMachineStateBase* _CompileLiteralMatch(Regex* regex, StringView* source)
{
	// literal match, will be modified over time
	StringView literal_match = StringView_Slice(*source, 0, 0);

	bool end_reached = false;
	if (SingleCharacterQuantified(*source)) {
		end_reached = true;
		ptrdiff_t length = UTF8_NextCodepoint(source->source) - source->source;
		literal_match = StringView_Slice(*source, 0, length);
	}

	while (! end_reached) {
		if (SingleCharacterQuantified(*source)) {
			/*
			This break allows for regexes like "a?"
			It also checks for one utf8 character (e.g. not flags)
			If it wouldn't break, "var?" would be treated as an
			optional word, Optional[var] not as "va" + Optional[r]
			*/
			break;
		}

		// Still in a literal sequence?
		if (! _IsLiteralChar(source->source[0])) {
			// Literal sequence is over, break here
			break;
		}

		// Increment literal match length
		source->source++;
		source->length--;
		literal_match.length++;

		// Also terminate the loop if there are no more characters left to read
		end_reached |= source->length == 0;
	}

	return _NewAcceptState(regex, &literal_match, literal_match.length);
}

static RegexMachineStateBase* _CompileEscapedStuff(Regex* regex, StringView* source)
{
	if (source->length == 1) {
		// TODO: Set Error
		return NULL;
	}

	*source = StringView_Slice(*source, 1, source->length);
	switch (UTF8_CodepointLength(source->source)) {
		case 1:
		{
			char next = source->source[0];
			switch (next) {
			case '\\':
			case '(':
			case ')':
			case '{':
			case '}':
			case '[':
			case ']':
			case '+':
			case '*':
			case '?':
				return _NewAcceptState(regex, source, 1);
			}
			break;
		}
		case 2:
		case 3:
		case 4:
		case SIZE_MAX:
			return NULL;
	}

	// TODO: Set Error
	return NULL;
}

static RegexMachineStateBase* _CompileNewState(Regex* regex, StringView* source)
{
	switch (source->source[0]) {
	case '(':
		break;
	case '[':
		break;
	case '\\':
		return _CompileEscapedStuff(regex, source);
	default:
		return _CompileLiteralMatch(regex, source);
	}

	return NULL;
}

struct RegexStackEntry {
	// the parent container itself (group or option)
	RegexMachineStateBase* parent;

	// last compiled state, next one will be appended here
	RegexMachineStateBase* tail;
};

static int _PushIfGroup(DynamicArray* entry_stack, RegexMachineStateBase* new_state)
{
	if (new_state->type == REGEXMACHINESTATETYPE_GROUP) {
		struct RegexStackEntry* new_entry;
		if (DynamicArray_AppendEmpty(entry_stack, (void**) &new_entry)) {
			return ENOMEM;
		}

		new_entry->parent = new_state;
		new_entry->tail   = new_state;
	}

	return EXIT_SUCCESS;
}

void _RedirectOptionTails(RegexMachineStateOption* option, RegexMachineStateBase* new_tail)
{
	while (option->next_option != NULL) {

		RegexMachineStateBase* tail = option->base.next;
		while (tail->next != NULL) {
			tail = tail->next;
		}

		// Append
		tail->next = new_tail;

		option = option->next_option;
	}

	return;
}

static int _CompileIntoGroup(Regex* regex, DynamicArray* entries, StringView* source)
{
	struct RegexStackEntry* group_entry = DynamicArray_GetPointer(entries, entries->reserved - 1);
	RegexMachineStateGroup* group = (RegexMachineStateGroup*) group_entry->parent;

	if (source->source[0] == ')' && group->number != 0) {
		// End group
		RegexMachineStateGroup* group_end = Scratchpad_Reserve(&regex->machine_memory, sizeof(*group));
		if (group_end == NULL) {
			return ENOMEM;
		}

		RegexMachineStateGroup_Init(group_end, group->name, group->number);

		if (group->base.next->type == REGEXMACHINESTATETYPE_OPTION) {
			_RedirectOptionTails((RegexMachineStateOption*) group->base.next, &group_end->base);
			// Branching in group
		} else {
			// Just linear members in group
			group_entry->tail->next = &group_end->base;
		}

		DynamicArray_RemoveFast(entries, entries->reserved - 1);

		*source = StringView_Slice(*source, 1, source->length);

	} else if (source->source[0] == ')' && group->number == 0) {

		// Closing parenthesis for virtual zeroth capture group
		// TODO: Report error
		return EXIT_FAILURE;

	} else if (source->source[0] == '|') {
		// Add new option as 'next' of group, embed everything parsed in it

		RegexMachineStateOption* option;
		option = Scratchpad_Reserve(&regex->machine_memory, sizeof(*option));
		if (option == NULL) {
			return ENOMEM;
		}

		RegexMachineStateOption_Init(option);
		option->base.next = group->base.next;

		// Add to stack
		{
			struct RegexStackEntry* new_entry;
			if (DynamicArray_AppendEmpty(entries, (void**) &new_entry)) {
				return ENOMEM;
			}

			new_entry->parent = &option->base;
			new_entry->tail   = group_entry->tail;
			while (new_entry->tail->next != NULL) {
				new_entry->tail = new_entry->tail->next;
			}
		}

		// redirect group next
		group->base.next = &option->base;
		group_entry->tail = &option->base;

	} else {

		RegexMachineStateBase* new_state = _CompileNewState(regex, source);
		if (new_state == NULL) {
			return ENOMEM;
		}

		// Append to linked list
		group_entry->tail->next = new_state;
		group_entry->tail = new_state;

		int add_code = _PushIfGroup(entries, new_state);
		if (add_code != EXIT_SUCCESS) {
			return add_code;
		}
	}


	return EXIT_SUCCESS;
}

static int _CompileIntoOption(Regex* regex, DynamicArray* entries, StringView* source)
{
	struct RegexStackEntry* option_entry = DynamicArray_GetPointer(entries, entries->reserved - 1);


	if (source->source[0] == '|') {
		// New option
		RegexMachineStateOption* new_option;
		new_option = Scratchpad_Reserve(&regex->machine_memory, sizeof(*new_option));
		if (new_option == NULL) {
			return ENOMEM;
		}

		RegexMachineStateOption_Init(new_option);

		option_entry->parent = &new_option->base;
		option_entry->tail   = &new_option->base;

		*source = StringView_Slice(*source, 1, source->length);

	} else if (source->source[0] == ')') {
		// Can't parse state here, finish option
		DynamicArray_RemoveFast(entries, entries->reserved - 1);
	} else {
		// Append to current option
		RegexMachineStateBase* new_state = _CompileNewState(regex, source);

		option_entry->tail->next = new_state;
		option_entry->tail       = new_state;

		int add_code = _PushIfGroup(entries, new_state);
		if (add_code != EXIT_SUCCESS) {
			return add_code;
		}
	}

	return EXIT_SUCCESS;
}

static int _HaveQuantifier(StringView source)
{
	StringView quantifiers = StringView_FromString("+*?{");
	return StringView_Contains(quantifiers, StringView_Slice(source, 0, 1));
}

static int _CompileQuantifier(Regex* regex, DynamicArray* entries, StringView* source)
{
	struct RegexStackEntry* last_entry = DynamicArray_GetPointer(entries, DynamicArray_GetLength(entries) - 1);

	size_t min, max;
	size_t char_count = 1;

	RegexMachineStateBase* repeatable = last_entry->tail;
	RegexMachineStateBase* repeat_end = repeatable;
	RegexMachineStateBase* previous;

	previous = last_entry->parent;
	while (previous->next != repeatable) {
		if (previous->next->type == REGEXMACHINESTATETYPE_GROUP
			&&
			(((RegexMachineStateGroup*) previous->next)->number ==
			((RegexMachineStateGroup*) repeatable)->number)
		) {
			repeatable = previous->next;
			break;
		}

		previous = previous->next;
	}

	switch (source->source[0]) {
		case '?':
			min = 0;
			max = 1;
			break;
		case '+':
			min = 1;
			max = SIZE_MAX;
			break;
		case '*':
			min = 0;
			max = SIZE_MAX;
			break;
		case '{':
			// TODO: Implement
			return EXIT_FAILURE;
			break;
	}

	*source = StringView_Slice(*source, char_count, source->length);

	bool greedy = true;
	bool possessive = false;
	if (StringView_StartsWith(*source, StringView_FromString("?"))) {
		greedy = false;
		*source = StringView_Slice(*source, 1, source->length);
	} else if (StringView_StartsWith(*source, StringView_FromString("+"))) {
		possessive = true;
		*source = StringView_Slice(*source, 1, source->length);
	}


	RegexMachineStateRepeat* repeat_state;
	repeat_state = Scratchpad_Reserve(&regex->machine_memory, sizeof(*repeat_state));
	if (repeat_state == NULL) {
		return ENOMEM;
	}

	RegexMachineStateRepeat_Init(repeat_state, min, max, greedy, possessive, repeatable);

	// Insert before repeatable
	// redirect repeatable end
	repeat_end->next = &repeat_state->base;
	previous->next   = &repeat_state->base;

	last_entry->tail = &repeat_state->base;

	return EXIT_SUCCESS;
}

static int _CompileIntoEntries(Regex* regex, DynamicArray* entries, StringView* source)
{
	struct RegexStackEntry* current = DynamicArray_GetPointer(entries, entries->reserved - 1);


	if (source->length != 0) {
		// Happily compile on
		if (_HaveQuantifier(*source)) {
			return _CompileQuantifier(regex, entries, source);
		}

		switch (current->parent->type) {

		case REGEXMACHINESTATETYPE_GROUP:
			return _CompileIntoGroup(regex, entries, source);

		case REGEXMACHINESTATETYPE_OPTION:
			return _CompileIntoOption(regex, entries, source);

		default:
			return EXIT_FAILURE;
		}

	} else {
		// Pop all options but not groups
		// (groups need closing parentheses)
	}

	return EXIT_SUCCESS;
}

int Internal_Regex_Compile(Regex* regex, StringView source)
{
	int return_code = EXIT_SUCCESS;
	DynamicArray parent_stack;
	if (DynamicArray_Create(&parent_stack, sizeof(struct RegexStackEntry), 4, regex->machine_memory.allocator)) {
		// unlikely
		return ENOMEM;
	}

	{
		// Create group zero
		// Spans entire match
		RegexMachineStateGroup* zero_group;
		zero_group = Scratchpad_Reserve(&regex->machine_memory, sizeof(*zero_group));
		if (zero_group == NULL) {
			// unlikely
			return_code = ENOMEM;
			goto defer_stack;
		}

		RegexMachineStateGroup_Init(zero_group, NULL, 0);

		regex->head = &zero_group->base;

		struct RegexStackEntry* zero_stackentry;
		// Can't fail, first push with sufficient capacity
		DynamicArray_AppendEmpty(&parent_stack, (void**) &zero_stackentry);
		zero_stackentry->parent = &zero_group->base;
		zero_stackentry->tail   = &zero_group->base;
	}

	while (source.length || (parent_stack.reserved != 1)) {
		int compile_code = _CompileIntoEntries(regex, &parent_stack, &source);

		if (compile_code != EXIT_SUCCESS) {
			return_code = compile_code;
			goto defer_stack;
		}
	}

	{
		// End group zero
		// Spans entire match
		RegexMachineStateGroup* zero_group_end;
		zero_group_end = Scratchpad_Reserve(&regex->machine_memory, sizeof(*zero_group_end));
		if (zero_group_end == NULL) {
			// unlikely
			return_code = ENOMEM;
			goto defer_stack;
		}

		RegexMachineStateGroup_Init(zero_group_end, NULL, 0);


		struct RegexStackEntry* zero_stackentry = DynamicArray_GetPointer(&parent_stack, 0);
		zero_stackentry->tail->next = &zero_group_end->base;
		zero_stackentry->tail = &zero_group_end->base;
	}

defer_stack:
	DynamicArray_Destroy(&parent_stack);

	return return_code;
}
